/*
 * Syncdaemon API
 *
 * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * Copyright 2010 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "config.h"
#include "syncdaemon-shares-interface.h"
#include "utils.h"

G_DEFINE_TYPE(SyncdaemonSharesInterface, syncdaemon_shares_interface, SYNCDAEMON_TYPE_INTERFACE)

struct _SyncdaemonSharesInterfacePrivate {
	GObject *proxy;
};

static void
syncdaemon_shares_interface_finalize (GObject *object)
{
	SyncdaemonSharesInterface *interface = SYNCDAEMON_SHARES_INTERFACE (object);

	if (interface->priv != NULL) {
		g_free (interface->priv);
	}

	G_OBJECT_CLASS (syncdaemon_shares_interface_parent_class)->finalize (object);
}

static void
syncdaemon_shares_interface_class_init (SyncdaemonSharesInterfaceClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = syncdaemon_shares_interface_finalize;
}

static void
share_created_cb (DBusGProxy *proxy, GHashTable *share_info, gpointer user_data)
{
	SyncdaemonDaemon *daemon = NULL;
	SyncdaemonSharesInterface *interface = SYNCDAEMON_SHARES_INTERFACE (user_data);

	g_object_get (G_OBJECT (interface), "daemon", &daemon, NULL);
	if (daemon != NULL)
		g_signal_emit_by_name (daemon, "share_created", TRUE, share_info);
}

static void
share_create_error_cb (DBusGProxy *proxy, GHashTable *share_info, const gchar *error, gpointer user_data)
{
	SyncdaemonDaemon *daemon = NULL;
	SyncdaemonSharesInterface *interface = SYNCDAEMON_SHARES_INTERFACE (user_data);

	g_object_get (G_OBJECT (interface), "daemon", &daemon, NULL);
	if (daemon != NULL)
		g_signal_emit_by_name (daemon, "share_created", FALSE, share_info);
}

static void
share_deleted_cb (DBusGProxy *proxy, GHashTable *share_info, gpointer user_data)
{
	SyncdaemonDaemon *daemon = NULL;
	SyncdaemonSharesInterface *interface = SYNCDAEMON_SHARES_INTERFACE (user_data);

	g_object_get (G_OBJECT (interface), "daemon", &daemon, NULL);
	if (daemon != NULL)
		g_signal_emit_by_name (daemon, "share_deleted", TRUE, share_info);
}

static void
share_delete_error_cb (DBusGProxy *proxy, GHashTable *share_info, const gchar *error, gpointer user_data)
{
	SyncdaemonDaemon *daemon = NULL;
	SyncdaemonSharesInterface *interface = SYNCDAEMON_SHARES_INTERFACE (user_data);

	g_object_get (G_OBJECT (interface), "daemon", &daemon, NULL);
	if (daemon != NULL)
		g_signal_emit_by_name (daemon, "share_deleted", FALSE, share_info);
}

static void
syncdaemon_shares_interface_init (SyncdaemonSharesInterface *interface)
{
	interface->priv = g_new0 (SyncdaemonSharesInterfacePrivate, 1);

	/* Setup DBus proxy */
	interface->priv->proxy = syncdaemon_interface_setup_proxy (SYNCDAEMON_INTERFACE (interface),
								   "com.ubuntuone.SyncDaemon",
								   "/shares", "com.ubuntuone.SyncDaemon.Shares");
	if (interface->priv->proxy != NULL) {
		/* Connect to DBus signals */
		dbus_g_proxy_add_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareCreated",
					 dbus_g_type_get_map ("GHashTable",
							      G_TYPE_STRING,
							      G_TYPE_STRING),
					 G_TYPE_INVALID);
		dbus_g_proxy_connect_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareCreated",
					     G_CALLBACK (share_created_cb), interface, NULL);

		dbus_g_proxy_add_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareCreateError",
					 dbus_g_type_get_map ("GHashTable",
							      G_TYPE_STRING,
							      G_TYPE_STRING),
					 G_TYPE_STRING,
					 G_TYPE_INVALID);
		dbus_g_proxy_connect_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareCreateError",
					     G_CALLBACK (share_create_error_cb), interface, NULL);

		dbus_g_proxy_add_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareDeleted",
					 dbus_g_type_get_map ("GHashTable",
							      G_TYPE_STRING,
							      G_TYPE_STRING),
					 G_TYPE_INVALID);
		dbus_g_proxy_connect_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareDeleted",
					     G_CALLBACK (share_deleted_cb), interface, NULL);

		dbus_g_proxy_add_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareDeleteError",
					 dbus_g_type_get_map ("GHashTable",
							      G_TYPE_STRING,
							      G_TYPE_STRING),
					 G_TYPE_STRING,
					 G_TYPE_INVALID);
		dbus_g_proxy_connect_signal (DBUS_G_PROXY (interface->priv->proxy), "ShareDeleteError",
					     G_CALLBACK (share_delete_error_cb), interface, NULL);
	}
}

/**
 * syncdaemon_shares_interface_new:
 */
SyncdaemonSharesInterface *
syncdaemon_shares_interface_new (SyncdaemonDaemon *daemon)
{
	g_return_val_if_fail (SYNCDAEMON_IS_DAEMON (daemon), NULL);

	return g_object_new (SYNCDAEMON_TYPE_SHARES_INTERFACE, "daemon", daemon, NULL);
}

/**
 * syncdaemon_shares_interface_accept:
 */
void
syncdaemon_shares_interface_accept (SyncdaemonSharesInterface *interface, const gchar *share_id)
{
	g_return_if_fail (SYNCDAEMON_IS_SHARES_INTERFACE (interface));
	g_return_if_fail (share_id != NULL);

	dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "accept_share",
				 no_output_dbus_call_ended_cb, interface, NULL,
				 G_TYPE_STRING, share_id,
				 G_TYPE_INVALID);
}

/**
 * syncdaemon_shares_interface_create:
 */
void
syncdaemon_shares_interface_create (SyncdaemonSharesInterface *interface,
				    const gchar *path,
				    GSList *usernames,
				    const gchar *name,
				    gboolean allow_modifications)
{
	g_return_if_fail (SYNCDAEMON_IS_SHARES_INTERFACE (interface));
	g_return_if_fail (path != NULL);
	g_return_if_fail (usernames != NULL);
	g_return_if_fail (name != NULL);

	if (g_slist_length (usernames) == 1) {
		dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "create_share",
					 no_output_dbus_call_ended_cb, interface, NULL,
					 G_TYPE_STRING, path,
					 G_TYPE_STRING, (const gchar *) usernames->data,
					 G_TYPE_STRING, name,
					 G_TYPE_STRING, allow_modifications ? "Modify" : "View",
					 G_TYPE_INVALID);
	} else {
		GSList *l;
		gint i;
		gchar **users_array = g_new0 (gchar *, g_slist_length (usernames));

		for (l = usernames, i = 0; l != NULL; l = l->next, i++)
			users_array[i] = g_strdup (l->data);

		dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "create_shares",
					 no_output_dbus_call_ended_cb, interface, NULL,
					 G_TYPE_STRING, path,
					 G_TYPE_STRV, users_array,
					 G_TYPE_STRING, name,
					 G_TYPE_STRING, allow_modifications ? "Modify" : "View",
					 G_TYPE_INVALID);

		g_strfreev (users_array);
	}
}

/**
 * syncdaemon_shares_interface_delete:
 */
void
syncdaemon_shares_interface_delete (SyncdaemonSharesInterface *interface, const gchar *share_id)
{
	g_return_if_fail (SYNCDAEMON_IS_SHARES_INTERFACE (interface));
	g_return_if_fail (share_id != NULL);

	dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "delete_share",
				 no_output_dbus_call_ended_cb, interface, NULL,
				 G_TYPE_STRING, share_id,
				 G_TYPE_INVALID);
}

static void
got_shared_cb (DBusGProxy *proxy, DBusGProxyCall *call_id, gpointer user_data)
{
	GSList *shared_list;
	SyncdaemonSharesInterface *interface = SYNCDAEMON_SHARES_INTERFACE (user_data);
	SyncdaemonDaemon *daemon = NULL;
	GError *error = NULL;

	g_object_get (G_OBJECT (interface), "daemon", &daemon, NULL);

	if (!dbus_g_proxy_end_call (proxy, call_id, &error,
				    dbus_g_type_get_collection ("GSList",
								dbus_g_type_get_map (
									"GHashTable",
									G_TYPE_STRING,
									G_TYPE_STRING)),
				    &shared_list,
				    G_TYPE_INVALID)) {
		if (daemon != NULL) {
			g_signal_emit_by_name (daemon, "error", error->message, NULL);
			g_error_free (error);

			return;
		}
	}

	while (shared_list != NULL) {
		GHashTable *share_info = shared_list->data;

		share_created_cb (proxy, share_info, interface);

		shared_list = g_slist_remove (shared_list, share_info);
		g_hash_table_unref (share_info);
	}
}

/**
 * syncdaemon_shares_interface_get_shared:
 */
void
syncdaemon_shares_interface_get_shared (SyncdaemonSharesInterface *interface)
{
	g_return_if_fail (SYNCDAEMON_IS_SHARES_INTERFACE (interface));

	dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "get_shared",
				 got_shared_cb, interface, NULL,
				 G_TYPE_INVALID);
}

/**
 * syncdaemon_shares_interface_refresh:
 */
void
syncdaemon_shares_interface_refresh (SyncdaemonSharesInterface *interface)
{
	g_return_if_fail (SYNCDAEMON_IS_SHARES_INTERFACE (interface));

	dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "refresh_shares",
				 no_output_dbus_call_ended_cb, interface, NULL,
				 G_TYPE_INVALID);
}

/**
 * syncdaemon_shares_interface_reject:
 */
void
syncdaemon_shares_interface_reject (SyncdaemonSharesInterface *interface, const gchar *share_id)
{
	g_return_if_fail (SYNCDAEMON_IS_SHARES_INTERFACE (interface));
	g_return_if_fail (share_id != NULL);

	dbus_g_proxy_begin_call (DBUS_G_PROXY (interface->priv->proxy), "reject_share",
				 no_output_dbus_call_ended_cb, interface, NULL,
				 G_TYPE_STRING, share_id,
				 G_TYPE_INVALID);
}
