/*
 * UbuntuOne Nautilus plugin
 *
 * Authors: Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * Copyright 2009-2010 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <glib/gi18n.h>
#include <libsyncdaemon/syncdaemon-folders-interface.h>
#include <libsyncdaemon/syncdaemon-shares-interface.h>
#include "location-widget.h"

G_DEFINE_TYPE(LocationWidget, location_widget, GTK_TYPE_HBOX)

static void
set_label_text (LocationWidget *location, gboolean enabled)
{
	gchar *labeltext;

	if (enabled) {
		labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s", _("folder synchronized"));
		gtk_widget_hide (location->help_label);
	} else {
		labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s", _("folder not synchronized"));
		gtk_widget_show (location->help_label);
	}
		
	gtk_label_set_markup (GTK_LABEL (location->info_label), labeltext);
	g_free (labeltext);
}

static void
start_spinner (LocationWidget *location)
{
	gtk_widget_hide (location->help_label);
	gtk_widget_show (location->spinner);
	gtk_spinner_start (GTK_SPINNER (location->spinner));
}

static void
stop_spinner (LocationWidget *location)
{
	gtk_spinner_stop (GTK_SPINNER (location->spinner));
	gtk_widget_hide (location->spinner);
}

static void
folder_created_cb (SyncdaemonDaemon *daemon, gboolean success, SyncdaemonFolderInfo *folder_info, gpointer user_data)
{
	const gchar *path;
	LocationWidget *location = LOCATION_WIDGET (user_data);

	path = syncdaemon_folder_info_get_path (folder_info);
	g_warning ("in folder_created_cb with path = %s", path);
	if (g_strcmp0 (path, location->path) == 0) {
		set_label_text (location, success);
		if (success) {
			gtk_widget_hide (location->enable_button);
			gtk_widget_show (location->disable_button);
			gtk_widget_set_sensitive (location->disable_button, TRUE);
			gtk_widget_hide (location->help_label);
		} else {
			gtk_widget_show (location->enable_button);
			gtk_widget_set_sensitive (location->enable_button, TRUE);
			gtk_widget_hide (location->disable_button);
			ubuntuone_show_error_dialog (location->uon, _("Error enabling folder"),
						     _("Could not enable folder %s for synchronizing to Ubuntu One"),
						     location->path);
		}

		stop_spinner (location);
	}
}

static void
folder_deleted_cb (SyncdaemonDaemon *daemon, gboolean success, SyncdaemonFolderInfo *folder_info, gpointer user_data)
{
	const gchar *path;
	LocationWidget *location = LOCATION_WIDGET (user_data);

	path = syncdaemon_folder_info_get_path (folder_info);
	if (g_strcmp0 (path, location->path) == 0) {
		if (success) {
			set_label_text (location, FALSE);
			gtk_widget_show (location->enable_button);
			gtk_widget_set_sensitive (location->enable_button, TRUE);
			gtk_widget_hide (location->disable_button);
			gtk_widget_show (location->help_label);
		} else {
			set_label_text (location, TRUE);
			gtk_widget_hide (location->enable_button);
			gtk_widget_show (location->disable_button);
			gtk_widget_set_sensitive (location->disable_button, TRUE);
			ubuntuone_show_error_dialog (location->uon, _("Error disabling folder"),
						     _("Could not disable folder %s for synchronizing to Ubuntu One"),
						     location->path);
		}

		stop_spinner (location);
	}
}

static void
location_widget_finalize (GObject *object)
{
	LocationWidget *location = LOCATION_WIDGET (object);

	g_signal_handlers_disconnect_by_func (location->uon->syncdaemon, folder_created_cb, location);
	g_signal_handlers_disconnect_by_func (location->uon->syncdaemon, folder_deleted_cb, location);

	if (location->path != NULL)
		g_free (location->path);

	G_OBJECT_CLASS (location_widget_parent_class)->finalize (object);
}

static void
location_widget_class_init (LocationWidgetClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);

	object_class->finalize = location_widget_finalize;
}

static void
disable_u1_cb (GtkButton *button, gpointer user_data)
{
	SyncdaemonInterface *interface;
	LocationWidget *location = LOCATION_WIDGET (user_data);

	/* Perform the removal of this folder */
	interface = syncdaemon_daemon_get_folders_interface (location->uon->syncdaemon);
	if (interface != NULL) {
		SyncdaemonFolderInfo *folder_info;

		folder_info = syncdaemon_folders_interface_get_info (
			SYNCDAEMON_FOLDERS_INTERFACE (interface),
			location->path);
		if (folder_info != NULL) {
			set_label_text (location, FALSE);
			gtk_widget_hide (location->disable_button);
			gtk_widget_show (location->enable_button);
			gtk_widget_set_sensitive (location->enable_button, FALSE);

			syncdaemon_folders_interface_delete (
				SYNCDAEMON_FOLDERS_INTERFACE (interface),
				syncdaemon_folder_info_get_volume_id (folder_info));

			g_object_unref (G_OBJECT (folder_info));
		}
	}
}

static void
enable_u1_cb (GtkButton *button, gpointer user_data)
{
	SyncdaemonInterface *interface;
	LocationWidget *location = LOCATION_WIDGET (user_data);

	/* Perform the addition of this folder */
	interface = syncdaemon_daemon_get_folders_interface (location->uon->syncdaemon);
	if (interface != NULL) {
		gchar *labeltext;

		/* Update label to show progress */
		labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s", _("sync in progress"));
		gtk_label_set_markup (GTK_LABEL (location->info_label), labeltext);
		g_free (labeltext);

		start_spinner (location);		
		gtk_widget_set_sensitive (location->enable_button, FALSE);

		syncdaemon_folders_interface_create (SYNCDAEMON_FOLDERS_INTERFACE (interface),
						     location->path);
	}
}

static void
location_widget_init (LocationWidget *location)
{
	GtkWidget *table, *box;

	location->path = NULL;
	location->info_label = location->enable_button = NULL;

	/* Create the widgets in the interface */
	table = gtk_table_new (3, 2, FALSE);
	gtk_widget_show (table);
	gtk_box_pack_start (GTK_BOX (location), table, TRUE, TRUE, 0);

	location->info_label = gtk_label_new ("<b>Ubuntu One</b>");
	gtk_label_set_use_markup (GTK_LABEL (location->info_label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (location->info_label), 0.0, 0.5);
	gtk_widget_show (location->info_label);
	gtk_table_attach (GTK_TABLE (table), location->info_label, 0, 1, 0, 1,
			  GTK_EXPAND | GTK_FILL, GTK_FILL, 3, 3);

	box = gtk_vbox_new (FALSE, 3);
	gtk_widget_show (box);
	gtk_table_attach (GTK_TABLE (table), box, 0, 1, 1, 2,
			  GTK_FILL | GTK_EXPAND, GTK_FILL, 3, 3);

	location->help_label = gtk_label_new (_("These files are not backed up and will not be available in your other computers"));
	gtk_label_set_use_markup (GTK_LABEL (location->help_label), TRUE);
	gtk_misc_set_alignment (GTK_MISC (location->help_label), 0.0, 0.5);
	gtk_box_pack_start (GTK_BOX (box), location->help_label, TRUE, TRUE, 0);

	location->spinner = gtk_spinner_new ();
	gtk_table_attach (GTK_TABLE (table), location->spinner, 1, 2, 0, 2,
			  GTK_SHRINK, GTK_SHRINK, 3, 3);

	/* Buttons */
	box = gtk_vbox_new (FALSE, 3);
	gtk_widget_show (box);
	gtk_table_attach (GTK_TABLE (table), box, 2, 3, 0, 2,
			  GTK_FILL, GTK_FILL, 3, 3);

	location->disable_button = gtk_button_new_with_label (_("Cancel Sync"));
	g_signal_connect (G_OBJECT (location->disable_button), "clicked",
			  G_CALLBACK (disable_u1_cb), location);
	gtk_box_pack_start (GTK_BOX (box), location->disable_button, FALSE, FALSE, 0);

	location->enable_button = gtk_button_new_with_label (_("Sync now"));
	g_signal_connect (G_OBJECT (location->enable_button), "clicked",
			  G_CALLBACK (enable_u1_cb), location);
	gtk_box_pack_start (GTK_BOX (box), location->enable_button, FALSE, FALSE, 0);
}

static gboolean
is_special_udf (UbuntuOneNautilus *uon, const gchar *path)
{
	gint i;
	gchar *special_udfs[] = {
		"Ubuntu One",
		"Ubuntu One/Shared With Me",
		".ubuntuone/Purchased from Ubuntu One",
		".local/share/ubuntuone"
	};

	for (i = 0; i < G_N_ELEMENTS (special_udfs); i++) {
		gchar *tmp;
		gboolean done = FALSE;

		tmp = g_build_filename (g_get_home_dir (), special_udfs[i], NULL);
		if (g_str_has_prefix (path, tmp))
			done = TRUE;

		g_free (tmp);

		if (done)
			return TRUE;
	}

	return FALSE;
}

GtkWidget *
location_widget_new (UbuntuOneNautilus *uon, const gchar *path)
{
	LocationWidget *location;
	gboolean is_root;

	g_return_val_if_fail (uon != NULL, NULL);
	g_return_val_if_fail (path != NULL, NULL);

	/* We only show the location bar for user's folders */
	if (!g_str_has_prefix (path, g_get_home_dir ()) ||
	    strlen (path) == strlen (g_get_home_dir ()))
		return NULL;

	location = (LocationWidget *) g_object_new (TYPE_LOCATION_WIDGET, NULL);
	location->uon = uon;
	location->path = g_strdup (path);

	/* Connect to Syncdaemon signals */
	g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_created",
			  G_CALLBACK (folder_created_cb), location);
	g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_deleted",
			  G_CALLBACK (folder_deleted_cb), location);

	if (ubuntuone_is_storagefs (uon, path, &is_root)) {
		/* Create label and disable button */
		set_label_text (location, TRUE);
		if (is_root && !is_special_udf (uon, path))
			gtk_widget_show (location->disable_button);
	} else {
		/* Check if this is the 'Shared with me' folder */
		if (is_special_udf (uon, path)) {
			set_label_text (location, TRUE);
		} else {
			GSList *udfs, *l;
			SyncdaemonInterface *interface;
			gboolean allow_enabling = TRUE;

			/* Check to see if this is the parent of a UDF */
			interface = syncdaemon_daemon_get_folders_interface (uon->syncdaemon);
			udfs = syncdaemon_folders_interface_get_folders (SYNCDAEMON_FOLDERS_INTERFACE (interface));
			for (l = udfs; l != NULL; l = l->next) {
				SyncdaemonFolderInfo *folder_info = SYNCDAEMON_FOLDER_INFO (l->data);

				if (g_str_has_prefix (syncdaemon_folder_info_get_path (folder_info), path))
					allow_enabling = FALSE;
			}

			g_slist_free (udfs);

			if (allow_enabling) {
				set_label_text (location, FALSE);
				gtk_widget_show (location->enable_button);
			} else {
				gchar *labeltext;

				labeltext = g_strdup_printf ("<b>Ubuntu One</b> - %s",
							     _("cannot synchronize this folder"));
				gtk_label_set_markup (GTK_LABEL (location->info_label), labeltext);
				g_free (labeltext);

				gtk_label_set_text (GTK_LABEL (location->help_label),
						    _("This folder cannot be synchronized because it contains "
						      "one or more folders that already synchronized"));
				gtk_widget_show (location->help_label);
			}
		}
	}

	gtk_widget_show (GTK_WIDGET (location));

	return GTK_WIDGET (location);
}
