/*
 * ubuntuone-nautilus.c - Nautilus extensions for Ubuntu One
 *
 * Authors: Tim Cole <tim.cole@canonical.com>
 *          Rodney Dawes <rodney.dawes@canonical.com>
 *          Rodrigo Moya <rodrigo.moya@canonical.com>
 *
 * Copyright 2009-2010 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n-lib.h>
#include <libnautilus-extension/nautilus-extension-types.h>
#include <libnautilus-extension/nautilus-file-info.h>
#include <libnautilus-extension/nautilus-info-provider.h>
#include <libnautilus-extension/nautilus-menu-provider.h>
#include <libnautilus-extension/nautilus-location-widget-provider.h>
#include <libsyncdaemon/libsyncdaemon.h>
#include "ubuntuone-nautilus.h"
#include "location-widget.h"
#include "u1-contacts-picker.h"

/* We need to do this explicitly because older versions of nautilus
 * don't seem to do it for us
 */
#include <gtk/gtk.h>

#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>

/* The header is generated from ubuntuone-marshallers.list */
#include "ubuntuone-marshallers.h"

/* This is used when a change is pending */
#define UPDATE_PENDING "pending"

static void ubuntuone_nautilus_finalize(GObject * object);

static void ubuntuone_nautilus_register_type (GTypeModule * module);

/* Utility functions */
static void ubuntuone_nautilus_reset_emblem (UbuntuOneNautilus * uon,
					     const char * path);

static void ubuntuone_nautilus_observed_file_unref (gpointer user_data,
                                                    GObject *where_the_object_was);

static void ubuntuone_nautilus_observed_file_foreach_unref (gpointer key,
							    gpointer value,
							    gpointer user_data);

/* DBus signal and async method call handlers */
static void ubuntuone_nautilus_update_meta (SyncdaemonFilesystemInterface *interface,
					    gboolean success,
					    SyncdaemonMetadata *metadata,
					    gpointer user_data);
static void ubuntuone_nautilus_status_changed (SyncdaemonDaemon *daemon,
					       SyncdaemonStatusInfo *status_info,
					       gpointer user_data);
static void ubuntuone_nautilus_upload_started (SyncdaemonDaemon *daemon,
					       gchar * path,
					       gpointer user_data);
static void ubuntuone_nautilus_upload_finished (SyncdaemonDaemon *daemon,
						 gchar * path,
						SyncdaemonTransferInfo * info,
						 gpointer user_data);
static void ubuntuone_nautilus_download_started (SyncdaemonDaemon *daemon,
						 gchar * path,
						 gpointer user_data);
static void ubuntuone_nautilus_download_finished (SyncdaemonDaemon *daemon,
						  gchar * path,
						  SyncdaemonTransferInfo * info,
						  gpointer user_data);
static void ubuntuone_nautilus_udf_created (SyncdaemonDaemon *daemon,
					    gboolean success,
					    SyncdaemonFolderInfo *folder_info,
					    gpointer user_data);
static void ubuntuone_nautilus_udf_deleted (SyncdaemonDaemon *daemon,
					    gboolean success,
					    SyncdaemonFolderInfo *folder_info,
					    gpointer user_data);
static void ubuntuone_nautilus_file_published (SyncdaemonDaemon *daemon,
					       GHashTable * hash,
					       gpointer user_data);
static void ubuntuone_nautilus_got_public_files (SyncdaemonDaemon *daemon,
                                                 GPtrArray * files,
                                                 gpointer user_data);
static void ubuntuone_nautilus_share_created (SyncdaemonDaemon *daemon,
					      gboolean success,
					      GHashTable * hash,
					      gpointer user_data);

static GObjectClass * parent_class = NULL;

static void ubuntuone_nautilus_reset_emblem (UbuntuOneNautilus * uon,
					     const char * path) {

  NautilusFileInfo *file = g_hash_table_lookup (uon->observed, path);
  if (file)
    nautilus_file_info_invalidate_extension_info (file);
}

/* Called when NautilusFileInfo is unreferenced */
static void ubuntuone_nautilus_observed_file_unref (gpointer user_data,
                                                    GObject *where_the_object_was) {
  UbuntuOneNautilus * uon;
  NautilusFileInfo * file;
  gchar * path = NULL;

  uon  = UBUNTUONE_NAUTILUS(user_data);
  file = NAUTILUS_FILE_INFO(where_the_object_was);
  path = g_filename_from_uri (nautilus_file_info_get_uri (file), NULL, NULL);

  if (g_hash_table_lookup (uon->observed, path ))
    g_hash_table_remove (uon->observed, path);

}

/* Cleanup routine for weak reference callbacks */
static void ubuntuone_nautilus_observed_file_foreach_unref (gpointer key,
							    gpointer value,
							    gpointer user_data) {
  UbuntuOneNautilus * uon;
  NautilusFileInfo * file;

  uon = UBUNTUONE_NAUTILUS(user_data);
  file = NAUTILUS_FILE_INFO(value);

  g_object_weak_unref (G_OBJECT(file),
		       (GWeakNotify)ubuntuone_nautilus_observed_file_unref,
		       uon);

}

/* Are we in an Ubuntu One managed (root or udf) directory */
gboolean ubuntuone_is_storagefs (UbuntuOneNautilus * uon,
				 const char * path,
				 gboolean * is_root) {
  gboolean managed = FALSE;
  gchar * dirpath;
  gchar * shared;
  GSList * udfs, * l;
  SyncdaemonFoldersInterface *interface;

  *is_root = FALSE;

  if (!uon->managed)
    return FALSE;

  if (!path)
    return FALSE;

  /* Check the Ubuntu One directory */
  if (strcmp (path, uon->managed) == 0) {
    *is_root = TRUE;
    return TRUE;
  }

  shared = g_build_filename (uon->managed, "Shared With Me", NULL);
  if (strncmp (path, shared, strlen (shared)) == 0) {
    managed = FALSE;
    goto fsdone;
  }

  dirpath = g_strdup_printf ("%s/", uon->managed);
  if (strncmp (path, dirpath, strlen (dirpath)) == 0) {
    managed = TRUE;
    goto fsdone;
  }

  g_free (dirpath);

  /* Check for UDFs */
  interface = SYNCDAEMON_FOLDERS_INTERFACE (syncdaemon_daemon_get_folders_interface (uon->syncdaemon));
  if (interface != NULL) {
    udfs = syncdaemon_folders_interface_get_folders (SYNCDAEMON_FOLDERS_INTERFACE (interface));
    for (l = udfs; l != NULL && l->data != NULL; l = l->next) {
      const char *tmp_path;

      tmp_path = syncdaemon_folder_info_get_path (SYNCDAEMON_FOLDER_INFO (l->data));
      if (!tmp_path)
        continue;

      if (strcmp (path, tmp_path) == 0) {
	 managed = TRUE;
	 *is_root = TRUE;
	 break;
       }

       dirpath = g_strdup_printf ("%s/", (char *) tmp_path);
       if (strncmp (path, dirpath, strlen (dirpath)) == 0) {
	 managed = TRUE;
	 g_free (dirpath);
	 break;
       }

       g_free (dirpath);
    }
    g_slist_free (udfs);
  }

 fsdone:
  g_free (shared);
  return managed;
}

static void ubuntuone_nautilus_add_observed (UbuntuOneNautilus * uon,
					     gchar * path,
					     NautilusFileInfo * file) {
  g_object_weak_ref (G_OBJECT(file),
		     (GWeakNotify)ubuntuone_nautilus_observed_file_unref,
		     uon);
  g_hash_table_insert (uon->observed, g_strdup (path), file);
}

/* Update file info provider */
static NautilusOperationResult ubuntuone_nautilus_update_file_info (NautilusInfoProvider * provider,
								    NautilusFileInfo * file,
								    GClosure * update_complete,
								    NautilusOperationHandle ** handle) {
  UbuntuOneNautilus * uon;
  gboolean is_root;
  char * path = NULL;

  uon = UBUNTUONE_NAUTILUS(provider);

  /* If syncdaemon is not ready, do nothing */
  if (!syncdaemon_daemon_is_ready (uon->syncdaemon))
    return NAUTILUS_OPERATION_COMPLETE;

  path = g_filename_from_uri (nautilus_file_info_get_uri (file), NULL, NULL);

  if (path == NULL)
    return NAUTILUS_OPERATION_COMPLETE;

  if (!g_hash_table_lookup (uon->updated, path) &&
      !g_hash_table_lookup (uon->needsupdating, path) &&
      ubuntuone_is_storagefs (uon, path, &is_root)) {
    SyncdaemonFilesystemInterface *interface;

    /* Add the unsynchronized emblem anyway, and update later */
    ubuntuone_nautilus_add_observed (uon, path, file);
    nautilus_file_info_add_emblem (file, "ubuntuone-unsynchronized");

    interface = (SyncdaemonFilesystemInterface *) syncdaemon_daemon_get_filesystem_interface (uon->syncdaemon);
    if (interface != NULL) {
      syncdaemon_filesystem_interface_get_metadata_async (interface,
							  path,
							  (SyncdaemonGotMetadataFunc) ubuntuone_nautilus_update_meta,
							  uon);
      goto updating_meta;
    }
  }

  if (g_hash_table_lookup (uon->uploads, path) ||
      g_hash_table_lookup (uon->downloads, path)) {
    nautilus_file_info_add_emblem (file, "ubuntuone-updating");
    ubuntuone_nautilus_add_observed (uon, path, file);
  } else if (g_hash_table_lookup (uon->updated, path)) {
    nautilus_file_info_add_emblem (file, "ubuntuone-synchronized");
    ubuntuone_nautilus_add_observed (uon, path, file);
  } else if (g_hash_table_lookup (uon->needsupdating, path)) {
    nautilus_file_info_add_emblem (file, "ubuntuone-unsynchronized");
    ubuntuone_nautilus_add_observed (uon, path, file);
  }

  if (g_hash_table_lookup (uon->udfs, path)) {
    nautilus_file_info_add_emblem (file, "ubuntuone-enabled");
    ubuntuone_nautilus_add_observed (uon, path, file);
  }

  if (g_hash_table_lookup (uon->public, path)) {
    nautilus_file_info_add_emblem (file, "ubuntuone-public");
    ubuntuone_nautilus_add_observed (uon, path, file);
  }

  if (g_hash_table_lookup (uon->shares, path)) {
    nautilus_file_info_add_emblem (file, "shared");
    ubuntuone_nautilus_add_observed (uon, path, file);
  }

 updating_meta:
  g_free (path);

  return NAUTILUS_OPERATION_COMPLETE;
}

static void ubuntuone_nautilus_info_provider_iface_init (NautilusInfoProviderIface * iface) {
  iface->update_file_info = ubuntuone_nautilus_update_file_info;
}

/* LocationWidget provider */
static GtkWidget * ubuntuone_nautilus_get_location_widget(NautilusLocationWidgetProvider * provider,
							  const char * uri,
							  GtkWidget * parent) {
  UbuntuOneNautilus * uon;
  gchar *path;
  GtkWidget *location;

  uon = UBUNTUONE_NAUTILUS (provider);

  /* If syncdaemon is not ready, do nothing */
  if (!syncdaemon_daemon_is_ready (uon->syncdaemon))
    return NULL;

  path = g_filename_from_uri (uri, NULL, NULL);

  if (!path)
    return NULL;

  location = location_widget_new (uon, path);

  g_free (path);

  return location;
}

static void ubuntuone_nautilus_bar_provider_iface_init (NautilusLocationWidgetProviderIface * iface) {
  iface->get_widget = ubuntuone_nautilus_get_location_widget;
}

static void __cb_data_free (struct _CBData * data) {
  if (!data)
    return;

  data->uon = NULL;
  data->parent = NULL;

  data->user_picker = NULL;
  data->allow_mods = NULL;

  if (data->path) {
    g_free (data->path);
    data->path = NULL;
  }

  g_free (data);

  data = NULL;
}

static void
picker_selection_changed_cb (U1ContactsPicker *picker, gpointer user_data)
{
  GSList *selection;
  GtkWidget * dialog = (GtkWidget *) user_data;

  selection = u1_contacts_picker_get_selected_emails (picker);
  if (selection != NULL) {
    gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT, TRUE);
    u1_contacts_picker_free_selection_list (selection);
  } else
    gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT, FALSE);
}

/* Share on Ubuntu One dialog constructor */
static GtkWidget * ubuntuone_nautilus_share_dialog_construct (struct _CBData * data) {
  GtkWidget * dialog;
  GtkWidget * area, * table, * label;

  dialog = gtk_dialog_new ();
  gtk_window_set_title (GTK_WINDOW (dialog), _("Share on Ubuntu One"));
  gtk_window_set_transient_for (GTK_WINDOW (dialog), GTK_WINDOW (data->parent));
  gtk_window_set_destroy_with_parent (GTK_WINDOW (dialog), TRUE);
  gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
  gtk_dialog_add_buttons (GTK_DIALOG (dialog),
			  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			  (_("Share")), GTK_RESPONSE_ACCEPT,
			  NULL);
  gtk_dialog_set_default_response (GTK_DIALOG (dialog),
				   GTK_RESPONSE_ACCEPT);
  gtk_dialog_set_response_sensitive (GTK_DIALOG (dialog), GTK_RESPONSE_ACCEPT, FALSE);
  gtk_window_set_icon_name (GTK_WINDOW (dialog), "ubuntuone");

  area = gtk_dialog_get_content_area (GTK_DIALOG (dialog));

  table = gtk_table_new (3, 2, FALSE);
  gtk_table_set_row_spacings (GTK_TABLE (table), 12);
  gtk_table_set_col_spacings (GTK_TABLE (table), 6);
  gtk_container_set_border_width (GTK_CONTAINER (table), 7);
  gtk_widget_show (table);
  gtk_container_add (GTK_CONTAINER (area), table);

  data->user_picker = u1_contacts_picker_new ();
  g_signal_connect (G_OBJECT (data->user_picker), "selection-changed",
		    G_CALLBACK (picker_selection_changed_cb), dialog);
  gtk_table_attach (GTK_TABLE (table), data->user_picker, 0, 2, 0, 2,
		    GTK_FILL | GTK_EXPAND | GTK_SHRINK,
		    GTK_FILL | GTK_EXPAND | GTK_SHRINK, 3, 3);
  gtk_widget_show (data->user_picker);

  data->allow_mods = gtk_check_button_new_with_mnemonic (_("_Allow Modification"));
  gtk_table_attach (GTK_TABLE (table), data->allow_mods, 0, 2, 2, 3, GTK_FILL, GTK_FILL, 3, 3);
  gtk_widget_show (data->allow_mods);

  gtk_widget_set_size_request (dialog, 500, 450);

  return dialog;
}

/* Menu callbacks */
static void ubuntuone_nautilus_share_dialog_response (GtkDialog * dialog,
						      gint response,
						      gpointer user_data) {
  struct _CBData * data = (struct _CBData *) user_data;

  switch (response) {
  case GTK_RESPONSE_ACCEPT: {
    GSList *emails;
    SyncdaemonSharesInterface *interface;
    gboolean allow_mods = FALSE;

    emails = u1_contacts_picker_get_selected_emails (U1_CONTACTS_PICKER (data->user_picker));
    if (emails == NULL) {
      GtkWidget *error_dialog;

      error_dialog = gtk_message_dialog_new (GTK_WINDOW (dialog),
					     GTK_DIALOG_DESTROY_WITH_PARENT,
					     GTK_MESSAGE_ERROR,
					     GTK_BUTTONS_CLOSE,
					     _("You need to select at least one contact to share this folder with"));
      gtk_dialog_run (GTK_DIALOG (error_dialog));
      gtk_widget_destroy (error_dialog);

      return;
    }

    allow_mods = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (data->allow_mods));

    interface = SYNCDAEMON_SHARES_INTERFACE (syncdaemon_daemon_get_shares_interface (data->uon->syncdaemon));
    if (interface != NULL) {
	    syncdaemon_shares_interface_create (interface,
						data->path,
						emails,
						g_path_get_basename (data->path),
						allow_mods);
    }
 
    u1_contacts_picker_free_selection_list (emails);
  }
  default:
    gtk_widget_destroy (GTK_WIDGET (dialog));
    break;
  }
}

static void ubuntuone_nautilus_share_dialog_closed (GtkDialog * dialog,
						    gpointer user_data) {
  ubuntuone_nautilus_share_dialog_response (dialog, GTK_RESPONSE_CANCEL,
					    user_data);
}

static void ubuntuone_nautilus_public_meta (SyncdaemonFilesystemInterface *interface,
					    gboolean success,
					    SyncdaemonMetadata *metadata,
					    gpointer user_data) {
  struct _CBData * data = (struct _CBData *) user_data;
  GError * error = NULL;
  const gchar * share_id, * node_id;
  SyncdaemonInterface *public;

  if (!success) {
    g_warning ("ERROR: getting metadata for public file");
    return;
  }

  share_id = syncdaemon_metadata_get_share_id (metadata);
  node_id = syncdaemon_metadata_get_node_id (metadata);

  public = syncdaemon_daemon_get_publicfiles_interface (data->uon->syncdaemon);
  if (public != NULL) {
    syncdaemon_publicfiles_interface_change_public_access (SYNCDAEMON_PUBLICFILES_INTERFACE (public),
							   share_id, node_id, data->make_public);
  }
}

static void ubuntuone_nautilus_copy_public_url (NautilusMenuItem * item,
						gpointer user_data) {
  struct _CBData * data = (struct _CBData *) user_data;
  gchar * url;

  url = g_hash_table_lookup (data->uon->public, data->path);
  gtk_clipboard_set_text (gtk_clipboard_get(GDK_SELECTION_CLIPBOARD),
			  url, strlen (url));
  gtk_clipboard_store (gtk_clipboard_get(GDK_SELECTION_CLIPBOARD));
}

static void ubuntuone_nautilus_toggle_publicity (NautilusMenuItem * item,
						 gpointer user_data) {
  SyncdaemonFilesystemInterface *interface;
  struct _CBData * data = (struct _CBData *) user_data;

  interface = (SyncdaemonFilesystemInterface *) syncdaemon_daemon_get_filesystem_interface (data->uon->syncdaemon);
  if (interface != NULL) {
    syncdaemon_filesystem_interface_get_metadata_async (interface,
							data->path,
							(SyncdaemonGotMetadataFunc) ubuntuone_nautilus_public_meta,
							data);
  }

  g_hash_table_replace (data->uon->public, g_strdup (data->path), UPDATE_PENDING);
  ubuntuone_nautilus_reset_emblem (data->uon, data->path);
}

static void ubuntuone_nautilus_share_folder (NautilusMenuItem * item,
					     gpointer * user_data) {
  struct _CBData * data = (struct _CBData *) user_data;
  GtkWidget * dialog;

  dialog = ubuntuone_nautilus_share_dialog_construct (data);
  g_signal_connect (dialog, "close",
		    G_CALLBACK (ubuntuone_nautilus_share_dialog_closed), data);
  g_signal_connect (dialog, "response",
		    G_CALLBACK (ubuntuone_nautilus_share_dialog_response), data);

  gtk_widget_show (dialog);
}

static void ubuntuone_nautilus_unshare_folder (NautilusMenuItem * item,
					       gpointer * user_data) {
  struct _CBData * data = (struct _CBData *) user_data;
  gchar * share_id;
  SyncdaemonSharesInterface *interface;

  interface = (SyncdaemonSharesInterface *) syncdaemon_daemon_get_shares_interface (data->uon->syncdaemon);
  if (interface != NULL) {
    share_id = g_hash_table_lookup (data->uon->shares, data->path);
    syncdaemon_shares_interface_delete (interface, share_id);
  }
}

/* Menu provider */
static GList * ubuntuone_nautilus_get_menu_items (NautilusMenuProvider * provider,
						  GtkWidget * window,
						  GList * files) {
  UbuntuOneNautilus * uon;
  NautilusFileInfo * file;
  GList * items = NULL;
  gchar * path;
  gchar * item;
  gchar * homedir_path;
  gboolean is_managed, is_root, is_udf, is_public, is_shared, is_pending;
  gboolean is_inhome;
  struct _CBData * cb_data;

  uon = UBUNTUONE_NAUTILUS (provider);

  /* If syncdaemon is not ready, do nothing */
  if (!syncdaemon_daemon_is_ready (uon->syncdaemon))
    return NULL;

  is_managed = is_root = is_udf = is_public = is_shared = is_pending = FALSE;
  is_inhome = FALSE;

  if (g_list_length (files) != 1)
    return NULL;

  file = g_list_nth_data (files, 0);
  path = g_filename_from_uri (nautilus_file_info_get_uri (file), NULL, NULL);

  if (path == NULL)
    goto done;

  if (ubuntuone_is_storagefs (uon, path, &is_root))
    is_managed = TRUE;

  homedir_path = g_strdup_printf ("%s/", g_get_home_dir());
  if (strncmp (path, homedir_path, strlen (homedir_path)) == 0)
	is_inhome = TRUE;
  g_free (homedir_path);

  if ((item = g_hash_table_lookup (uon->udfs, path)) != NULL) {
    is_udf = TRUE;
	if (strcmp (item, UPDATE_PENDING) == 0)
	  is_pending = TRUE;
  } else if ((item = g_hash_table_lookup (uon->public, path)) != NULL) {
    is_public = TRUE;
	if (strcmp (item, UPDATE_PENDING) == 0)
	  is_pending = TRUE;
  }
  if (g_hash_table_lookup (uon->shares, path) != NULL)
    is_shared = TRUE;

  if (uon->cb_data)
    __cb_data_free (uon->cb_data);

  cb_data = g_new0 (struct _CBData, 1);
  cb_data->uon = uon;
  cb_data->parent = window;
  cb_data->path = g_strdup (path);

  uon->cb_data = cb_data;

  if ((is_managed || is_udf) && !is_root && nautilus_file_info_is_directory (file)) {
    NautilusMenuItem * item;

    item = nautilus_menu_item_new ("ubuntuone-share",
				   _("Share on Ubuntu One..."),
				   _("Share this folder on Ubuntu One"),
				   "ubuntuone");
	if (is_pending)
	  g_object_set (item, "sensitive", FALSE, NULL);

    g_signal_connect (item, "activate",
		      G_CALLBACK (ubuntuone_nautilus_share_folder),
		      cb_data);
    items = g_list_prepend (items, item);
  }

  if ((is_managed && is_shared) && !is_root &&  nautilus_file_info_is_directory (file)) {
    NautilusMenuItem * item;

    item = nautilus_menu_item_new ("ubuntuone-unshare",
				   _("Stop sharing on Ubuntu One..."),
				   _("Stop sharing this folder on Ubuntu One"),
				   "ubuntuone");
	if (is_pending)
	  g_object_set (item, "sensitive", FALSE, NULL);

    g_signal_connect (item, "activate",
		      G_CALLBACK (ubuntuone_nautilus_unshare_folder),
		      cb_data);
    items = g_list_prepend (items, item);
  }
    
  if (is_managed && (nautilus_file_info_get_file_type (file) == G_FILE_TYPE_REGULAR)) {
    NautilusMenuItem * item, * urlitem;

    if (is_public) {
      urlitem = nautilus_menu_item_new ("ubuntuone-geturl",
					_("Copy Ubuntu One public URL"),
					_("Copy the Ubuntu One public URL for this file to the clipboard."),
					"ubuntuone");
	  if (is_pending)
		g_object_set (urlitem, "sensitive", FALSE, NULL);

      g_signal_connect (urlitem, "activate",
			G_CALLBACK (ubuntuone_nautilus_copy_public_url),
			cb_data);
      items = g_list_append (items, urlitem);
      item = nautilus_menu_item_new ("ubuntuone-unpublish",
                                     _("Stop publishing via Ubuntu One"),
                                     _("No longer share this file with everyone."),
                                     "ubuntuone");
	  if (is_pending)
		g_object_set (item, "sensitive", FALSE, NULL);

      cb_data->make_public = FALSE;
    } else {
      item = nautilus_menu_item_new ("ubuntuone-publish",
                                     _("Publish via Ubuntu One"),
                                     _("Make this file available to anyone."),
                                     "ubuntuone");
      cb_data->make_public = TRUE;
    }
    g_signal_connect (item, "activate",
                      G_CALLBACK (ubuntuone_nautilus_toggle_publicity),
                      cb_data);
    items = g_list_append (items, item);
  }

 done:
  g_free (path);

  return items;
}

static GList * ubuntuone_nautilus_get_bg_menu_items (NautilusMenuProvider * provider,
						     GtkWidget * window,
						     NautilusFileInfo * folder) {
  GList * files = NULL;
  GList * items = NULL;

  files = g_list_prepend (files, folder);
  items = ubuntuone_nautilus_get_menu_items (provider, window, files);
  files = g_list_remove (files, folder);
  g_list_free (files);

  return items;
}

static void ubuntuone_nautilus_menu_provider_iface_init (NautilusMenuProviderIface * iface) {
  iface->get_file_items = ubuntuone_nautilus_get_menu_items;
  iface->get_background_items = ubuntuone_nautilus_get_bg_menu_items;
}

/* GType and nautilus module stuff */
static GType un_type = 0;

static void ubuntuone_nautilus_instance_init (UbuntuOneNautilus * uon) {
  SyncdaemonSharesInterface *interface;

  uon->connected = FALSE;
  uon->uploads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  uon->downloads = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  uon->shares = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  uon->updated = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  uon->needsupdating = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  uon->observed = g_hash_table_new_full (g_str_hash, g_str_equal, g_free, NULL);
  uon->udfs = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
  uon->public = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);

  uon->syncdaemon = syncdaemon_daemon_new ();

  /* Default to ~/Ubuntu One for now, as it's all we really support */
  uon->managed = g_build_filename (g_get_home_dir (), "Ubuntu One", NULL);
  uon->gotroot = FALSE;
  uon->gotudfs = FALSE;
  uon->gotpubs = FALSE;

  /* Connec to status-related signals */
  g_signal_connect (G_OBJECT (uon->syncdaemon), "status_changed",
		    G_CALLBACK (ubuntuone_nautilus_status_changed), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "upload_started",
		    G_CALLBACK (ubuntuone_nautilus_upload_started), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "upload_finished",
		    G_CALLBACK (ubuntuone_nautilus_upload_finished), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "download_started",
		    G_CALLBACK (ubuntuone_nautilus_download_started), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "download_finished",
		    G_CALLBACK (ubuntuone_nautilus_download_finished), uon);

  /* Connect to shares-related signals */
  g_signal_connect (G_OBJECT (uon->syncdaemon), "share_created",
		    G_CALLBACK (ubuntuone_nautilus_share_created), uon);

  /* Connect to folder-related signals */
  g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_created",
		    G_CALLBACK (ubuntuone_nautilus_udf_created), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_subscribed",
		    G_CALLBACK (ubuntuone_nautilus_udf_created), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_deleted",
		    G_CALLBACK (ubuntuone_nautilus_udf_deleted), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "folder_unsubscribed",
		    G_CALLBACK (ubuntuone_nautilus_udf_deleted), uon);

  /* Connect to public files-related signals */
  g_signal_connect (G_OBJECT (uon->syncdaemon), "file_published",
		    G_CALLBACK (ubuntuone_nautilus_file_published), uon);
  g_signal_connect (G_OBJECT (uon->syncdaemon), "got_published_files",
		    G_CALLBACK (ubuntuone_nautilus_got_public_files), uon);

}

static void ubuntuone_nautilus_class_init (UbuntuOneNautilusClass * klass) {
  parent_class = g_type_class_peek_parent (klass);

  G_OBJECT_CLASS(klass)->finalize = ubuntuone_nautilus_finalize;
}

static void ubuntuone_nautilus_finalize(GObject * object) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS(object);

  __cb_data_free (uon->cb_data);

  if (uon->syncdaemon)
    g_object_unref (uon->syncdaemon);

  g_hash_table_destroy (uon->uploads);
  uon->uploads = NULL;

  g_hash_table_destroy (uon->downloads);
  uon->downloads = NULL;

  g_hash_table_destroy (uon->shares);
  uon->shares = NULL;

  g_hash_table_destroy (uon->udfs);
  uon->udfs = NULL;

  g_hash_table_destroy (uon->public);
  uon->public = NULL;

  g_hash_table_destroy (uon->updated);
  uon->updated = NULL;

  g_hash_table_destroy (uon->needsupdating);
  uon->needsupdating = NULL;

  /* We need to remove all weak reference callbacks */
  g_hash_table_foreach (uon->observed, 
			(GHFunc)ubuntuone_nautilus_observed_file_foreach_unref,
			uon);

  g_hash_table_destroy (uon->observed);
  uon->observed = NULL;
}

GType ubuntuone_nautilus_get_type (void) {
  return un_type;
}

static void ubuntuone_nautilus_register_type (GTypeModule * module) {
  static const GTypeInfo info = {
    sizeof (UbuntuOneNautilusClass),
    (GBaseInitFunc) NULL,
    (GBaseFinalizeFunc) NULL,
    (GClassInitFunc) ubuntuone_nautilus_class_init,
    NULL,
    NULL,
    sizeof (UbuntuOneNautilus),
    0,
    (GInstanceInitFunc) ubuntuone_nautilus_instance_init,
  };

  static const GInterfaceInfo info_provider_iface_info = {
    (GInterfaceInitFunc) ubuntuone_nautilus_info_provider_iface_init,
    NULL,
    NULL
  };

  static const GInterfaceInfo bar_provider_iface_info = {
    (GInterfaceInitFunc) ubuntuone_nautilus_bar_provider_iface_init,
    NULL,
    NULL
  };

  static const GInterfaceInfo menu_provider_iface_info = {
    (GInterfaceInitFunc) ubuntuone_nautilus_menu_provider_iface_init,
    NULL,
    NULL
  };

  un_type = g_type_module_register_type (module, 
					 G_TYPE_OBJECT,
					 "UbuntuOneNautilus",
					 &info, 0);
  
  g_type_module_add_interface (module,
			       un_type,
			       NAUTILUS_TYPE_INFO_PROVIDER,
			       &info_provider_iface_info);

  g_type_module_add_interface (module,
			       un_type,
			       NAUTILUS_TYPE_LOCATION_WIDGET_PROVIDER,
			       &bar_provider_iface_info);

  /* XXX Need to sign a request via DBus */
  g_type_module_add_interface (module,
			       un_type,
			       NAUTILUS_TYPE_MENU_PROVIDER,
			       &menu_provider_iface_info);
}


/* DBus signal handlers and async method call handlers */
static void ubuntuone_nautilus_update_meta (SyncdaemonFilesystemInterface *interface,
					    gboolean success,
					    SyncdaemonMetadata *metadata,
					    gpointer user_data) {
  UbuntuOneNautilus * uon;
  GError * error = NULL;
  const gchar * local, * server, * path;
  gchar * new_path;

  if (!success) {
    g_warning ("ERROR: Could not retrieve metadata");
    return;
  }

  uon = UBUNTUONE_NAUTILUS (user_data);

  path = syncdaemon_metadata_get_path (metadata);
  local = syncdaemon_metadata_get_local_hash (metadata);
  server = syncdaemon_metadata_get_server_hash (metadata);

  new_path = g_strdup (path);

  gboolean is_updated = local && server &&
			strlen (local) > 0 &&
			strlen (server) > 0 &&
			strcmp (local, server) == 0;

  if (is_updated)
    g_hash_table_replace (uon->updated, new_path, new_path);
  else
    g_hash_table_replace (uon->needsupdating, new_path, new_path);
  ubuntuone_nautilus_reset_emblem (uon, new_path);
}

static void ubuntuone_nautilus_status_changed (SyncdaemonDaemon *daemon,
					       SyncdaemonStatusInfo *status_info,
					       gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  const gchar * status;
  gboolean is_online;

  is_online = syncdaemon_status_info_get_online (status_info);
  status = syncdaemon_status_info_get_name (status_info);

  /* Get the root when we get a status change signal, if we haven't yet */
  if (!uon->gotroot) {
    uon->managed = g_strdup (syncdaemon_daemon_get_root_dir (uon->syncdaemon));
    if (uon->managed != NULL)
      uon->gotroot = TRUE;
  }

  /* Get the list of UDFs if we haven't already */
  if (!uon->gotudfs) {
    SyncdaemonInterface *interface;

    interface = syncdaemon_daemon_get_folders_interface (uon->syncdaemon);
    if (interface != NULL) {
	    GSList *folders, *l;

      folders = syncdaemon_folders_interface_get_folders (SYNCDAEMON_FOLDERS_INTERFACE (interface));
      for (l = folders; l != NULL; l = l->next) {
        SyncdaemonFolderInfo *folder_info = SYNCDAEMON_FOLDER_INFO (l->data);

	ubuntuone_nautilus_udf_created (uon->syncdaemon, TRUE, folder_info, uon);
	uon->gotudfs = TRUE;
      }

      g_slist_free (folders);
    }
  }

  /* Get the list of public files if we haven't already */
  if (is_online && !uon->gotpubs) {
    SyncdaemonInterface *public;

    public = syncdaemon_daemon_get_publicfiles_interface (uon->syncdaemon);
    if (public != NULL) {
      syncdaemon_publicfiles_interface_get_public_files (SYNCDAEMON_PUBLICFILES_INTERFACE (public));
      uon->gotpubs = TRUE;
    }
  }
}

static void ubuntuone_nautilus_upload_started (SyncdaemonDaemon *daemon,
					       gchar * path,
					       gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);

  if (!g_hash_table_lookup (uon->uploads, path)) {
    gchar *new_path = g_strdup (path);
    g_hash_table_insert (uon->uploads, new_path, new_path);
    ubuntuone_nautilus_reset_emblem (uon, path);
  }
}

static void ubuntuone_nautilus_upload_finished (SyncdaemonDaemon *daemon,
						gchar * path,
						SyncdaemonTransferInfo * info,
						gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  gchar * new_path;

  g_hash_table_remove (uon->uploads, path);
  g_hash_table_remove (uon->needsupdating, path);

  new_path = g_strdup (path);
  g_hash_table_replace (uon->updated, new_path, new_path);

  ubuntuone_nautilus_reset_emblem (uon, path);
}

static void ubuntuone_nautilus_download_started (SyncdaemonDaemon *daemon,
						 gchar * path,
						 gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);

  if (!g_hash_table_lookup (uon->downloads, path)) {
    gchar *new_path = g_strdup (path);
    g_hash_table_insert (uon->downloads, new_path, new_path);
    ubuntuone_nautilus_reset_emblem (uon, path);
  }
}

static void ubuntuone_nautilus_download_finished (SyncdaemonDaemon *daemon,
						  gchar * path,
						  SyncdaemonTransferInfo * info,
						  gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  gchar * new_path;

  g_hash_table_remove (uon->downloads, path);
  g_hash_table_remove (uon->needsupdating, path);

  new_path = g_strdup (path);
  g_hash_table_replace (uon->updated, new_path, new_path);

  ubuntuone_nautilus_reset_emblem (uon, path);
}

static void ubuntuone_nautilus_udf_created (SyncdaemonDaemon *daemon,
					    gboolean success,
					    SyncdaemonFolderInfo *folder_info,
					    gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  const gchar * path, * id;
  gboolean subscribed;

  path = syncdaemon_folder_info_get_path (folder_info);
  id = syncdaemon_folder_info_get_volume_id (folder_info);
  subscribed = syncdaemon_folder_info_get_subscribed (folder_info);
  if (success) {
    if (!g_hash_table_lookup (uon->udfs, path)) {
      if (subscribed) {
	g_hash_table_replace (uon->udfs, g_strdup (path), g_strdup (id));
	ubuntuone_nautilus_reset_emblem (uon, path);
      }
    }
  }
}

static void ubuntuone_nautilus_udf_deleted (SyncdaemonDaemon *daemon,
					    gboolean success,
					    SyncdaemonFolderInfo *folder_info,
					    gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  const gchar * path;

  path = syncdaemon_folder_info_get_path (folder_info);
  if (path != NULL && success) {
    g_hash_table_remove (uon->udfs, path);
    ubuntuone_nautilus_reset_emblem (uon, path);
  }
}

static void ubuntuone_nautilus_file_published (SyncdaemonDaemon *daemon,
					       GHashTable *file_info,
					       gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  gchar * path, * url, * public;
  gboolean is_public;

  path = g_hash_table_lookup (file_info, "path");
  url = g_hash_table_lookup (file_info, "public_url");
  public = g_hash_table_lookup (file_info, "is_public");
  is_public = (public != NULL && strlen (public) != 0);

  if (!is_public && g_hash_table_lookup (uon->public, path))
    g_hash_table_remove (uon->public, path);

  if (is_public)
    g_hash_table_replace (uon->public, g_strdup (path), g_strdup (url));

  ubuntuone_nautilus_reset_emblem (uon, path);
  uon->gotpubs = TRUE;
}

static void ubuntuone_nautilus_got_public_files (SyncdaemonDaemon *daemon,
                                                 GPtrArray *files,
                                                 gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  int i;

  g_hash_table_remove_all (uon->public);
  for (i = 0; i < files->len; i++)
    g_hash_table_insert (uon->public,
                         g_strdup (g_hash_table_lookup (files->pdata[i], "path")),
                         g_strdup (g_hash_table_lookup (files->pdata[i], "public_url")));
}

static void ubuntuone_nautilus_share_created (SyncdaemonDaemon *daemon,
					      gboolean success,
					      GHashTable * hash,
					      gpointer user_data) {
  UbuntuOneNautilus * uon = UBUNTUONE_NAUTILUS (user_data);
  gchar * path;

  path = g_hash_table_lookup (hash, "path");
  if (success) {
    gchar * id;
    id = g_hash_table_lookup (hash, "volume_id");

    if (!g_hash_table_lookup (uon->shares, path)) {
      g_hash_table_insert (uon->shares, g_strdup (path), g_strdup (id));
      ubuntuone_nautilus_reset_emblem (uon, path);
    }
  } else {
    ubuntuone_show_error_dialog (uon, _("Error creating share."),
				 _("There was an error sharing the folder '%s'"),
				 path);
  }
}

/* Required Nautilus module handling methods */
void nautilus_module_initialize (GTypeModule * module) {
#ifdef ENABLE_NLS
  bindtextdomain (GETTEXT_PACKAGE, GNOMELOCALEDIR);
  bind_textdomain_codeset (GETTEXT_PACKAGE, "UTF-8");
  textdomain (GETTEXT_PACKAGE);
#endif

  ubuntuone_nautilus_register_type (module);
}

void nautilus_module_shutdown (void) {
}


void nautilus_module_list_types (const GType ** types,
				 int * num_types) {
  static GType type_list[1];
  
  type_list[0] = UBUNTUONE_TYPE_NAUTILUS;

  *types = type_list;
  *num_types = 1;
}
