#
# Author: Guillermo Gonzalez <guillermo.gonzalez@canonical.com>
#
# Copyright 2009 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License version 3, as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests for Sync."""

from __future__ import with_statement

import copy
import logging
import os
import shutil
import unittest
import uuid

import dbus
from dbus.mainloop.glib import DBusGMainLoop

from contrib.testing.testcase import (
    FakeVolumeManager,
    BaseTwistedTestCase,
    MementoHandler
)

from contrib.testing import testcase
from ubuntuone.syncdaemon.dbus_interface import DBusInterface
from ubuntuone.syncdaemon.filesystem_manager import FileSystemManager
from ubuntuone.syncdaemon.fsm import fsm as fsm_module
from ubuntuone.syncdaemon.main import Main
from ubuntuone.syncdaemon.sync import FSKey, Sync, SyncStateMachineRunner
from ubuntuone.syncdaemon.volume_manager import Share
from ubuntuone.syncdaemon.event_queue import EventQueue
from ubuntuone.storageprotocol.request import ROOT
from ubuntuone.storageprotocol import delta
from ubuntuone.syncdaemon.marker import MDMarker

DBusInterface.test = True

class FSKeyTestCase(BaseTwistedTestCase):
    """ Base test case for FSKey """

    def setUp(self):
        """ Setup the test """
        unittest.TestCase.setUp(self)
        self.test_dir = self.mktemp("test_dir")
        self.shares_dir = os.path.join(self.test_dir, 'shares')
        os.makedirs(self.shares_dir)
        self.root_dir = os.path.join(self.test_dir, 'root')
        os.makedirs(self.root_dir)
        self.fsmdir = os.path.join(self.test_dir, "fsmdir")
        self.partials_dir = os.path.join(self.test_dir, "partials")
        os.makedirs(self.partials_dir)
        self.fsm = FileSystemManager(self.fsmdir, self.partials_dir,
                                     FakeVolumeManager(self.root_dir))
        self.eq = EventQueue(self.fsm)
        self.fsm.register_eq(self.eq)
        self.share = self.create_share('share', 'share_name',
                                            self.fsm, self.shares_dir)
        self.share_path = self.share.path

    def tearDown(self):
        """ Clean up the test dir"""
        self.eq.shutdown()
        shutil.rmtree(self.test_dir)

    @staticmethod
    def create_share(share_id, share_name, fsm, shares_dir,
                     access_level='Modify'):
        """ creates a share """
        share_path = os.path.join(shares_dir, share_name)
        os.makedirs(share_path)
        share = Share(path=share_path, volume_id=share_id,
                      access_level=access_level)
        fsm.vm.add_share(share)
        return share


class FSKeyTests(FSKeyTestCase):
    """Test FSKey methods."""

    def test_get_mdid(self):
        """simple tests for get_mdid"""
        path = os.path.join(self.share.path, 'path')
        mdid = self.fsm.create(path, "share", node_id='uuid1')
        key = FSKey(self.fsm, path=path)
        self.assertEquals(mdid, key.get_mdid())
        key = FSKey(self.fsm, share_id='share', node_id='uuid1')
        self.assertEquals(mdid, key.get_mdid())
        key = FSKey(self.fsm, mdid=mdid)
        self.assertEquals(mdid, key.get_mdid())
        # with bad keys
        key = FSKey(self.fsm, share='share', node_id='uuid1')
        self.assertRaises(KeyError, key.get_mdid)
        # 1 valid and 1 invalid key
        key = FSKey(self.fsm, share='share', mdid=mdid)
        self.assertRaises(KeyError, key.get_mdid)

    def test_set(self):
        """test that changes to the key are keeped in _changes until sync"""
        path = os.path.join(self.share.path, 'path')
        self.fsm.create(path, "share", node_id='uuid1')
        key = FSKey(self.fsm, path=path)
        key.set(local_hash='a_hash')
        self.assertEquals('a_hash', key._changes['local_hash'])
        key.set(server_hash='a_hash_1')
        self.assertEquals('a_hash_1', key._changes['server_hash'])
        key.sync()
        self.assertEquals({}, key._changes)

    def test_sync(self):
        """test sync of the changes to the fsm"""
        path = os.path.join(self.share.path, 'path')
        mdid = self.fsm.create(path, "share", node_id='uuid1')
        key = FSKey(self.fsm, path=path)
        # change it
        key.set(local_hash='local_hash', server_hash='server_hash')
        # sync!
        key.sync()
        # get the mdobj and check the values
        mdobj = self.fsm.get_by_mdid(mdid)
        self.assertEquals('server_hash', mdobj.server_hash)
        self.assertEquals('local_hash', mdobj.local_hash)
        self.assertEquals('share', mdobj.share_id)
        self.assertEquals('uuid1', mdobj.node_id)

    def test_mdid_reset(self):
        """Test for mdid reset after a deletion"""
        # create a node
        path = os.path.join(self.share.path, 'path')
        self.fsm.create(path, "share", node_id='uuid1')
        key = FSKey(self.fsm, path=path)
        # fake a conflict and delete the metadata
        key.move_to_conflict()
        key.delete_metadata()
        # create the node again
        self.fsm.create(path, "share", node_id='uuid1')
        key.set(server_hash="")
        key.set(local_hash="")
        # fail no more!
        key.sync()
        # now check the delete_file method
        # fake a conflict and delete the metadata
        key.delete_file()
        # create the node again
        self.fsm.create(path, "share", node_id='uuid1')
        key.set(server_hash="")
        key.set(local_hash="")
        # fail no more!
        key.sync()

    def test_changes_reset(self):
        """Test for _changes reset after a deletion"""
        # create a node
        path = os.path.join(self.share.path, 'path')
        self.fsm.create(path, "share", node_id='uuid1')
        key = FSKey(self.fsm, path=path)
        # set a value and delete the metadata
        key.set(server_hash="server_hash")
        key.delete_metadata()
        # create the node again
        self.fsm.create(path, "share", node_id='uuid1')
        key.sync()
        self.assertNotEquals('server_hash',
                             self.fsm.get_by_path(path).server_hash)
        # now check the delete_file method
        key = FSKey(self.fsm, path=path)
        key.set(server_hash="server_hash1")
        # fake a conflict and delete the metadata
        key.delete_file()
        # create the node again
        self.fsm.create(path, "share", node_id='uuid1')
        key.sync()
        self.assertNotEquals('server_hash1',
                             self.fsm.get_by_path(path).server_hash)


class BaseSync(BaseTwistedTestCase):
    """Base test infrastructure for Sync."""

    def setUp(self):
        """Init."""
        BaseTwistedTestCase.setUp(self)
        self.root = self.mktemp('root')
        self.shares = self.mktemp('shares')
        self.data = self.mktemp('data')
        self.partials_dir = self.mktemp('partials_dir')
        self.handler = MementoHandler()
        self.handler.setLevel(logging.ERROR)
        self.main = Main(root_dir=self.root,
                         shares_dir=self.shares,
                         data_dir=self.data,
                         partials_dir=self.partials_dir,
                         host='localhost', port=0,
                         dns_srv=False, ssl=False,
                         disable_ssl_verify=True,
                         realm='fake.realm',
                         mark_interval=60,
                         handshake_timeout=2,
                         glib_loop=DBusGMainLoop(set_as_default=True))
        logging.getLogger('ubuntuone.SyncDaemon').addHandler(self.handler)
        dbus.service.BusName.__del__ = lambda _: None

    def tearDown(self):
        """Clean up."""
        self.main.shutdown()
        shutil.rmtree(self.root)
        shutil.rmtree(self.shares)
        shutil.rmtree(self.data)
        for record in self.handler.records:
            exc_info = getattr(record, 'exc_info', None)
            if exc_info is not None:
                raise exc_info[0], exc_info[1], exc_info[2]
        BaseTwistedTestCase.tearDown(self)


class TestSync(BaseSync):
    """Test for Sync."""

    def test_deleting_open_files_is_no_cause_for_despair(self):
        """test_deleting_open_files_is_no_cause_for_despair."""
        def cb(_):
            d0 = self.main.wait_for('HQ_HASH_NEW')
            f = file(self.root + '/a_file', 'w')
            print >> f, 'hola'
            os.unlink(self.root + '/a_file')
            f.close()
            print >> file(self.root + '/b_file', 'w'), 'chau'
            return d0
        d = self.main.wait_for('SYS_LOCAL_RESCAN_DONE')
        self.main.start()
        d.addCallback(cb)
        return d

    def test_stomp_deleted_file_is_no_cause_for_despair_either(self):
        """test_stomp_deleted_file_is_no_cause_for_despair_either."""
        def cb(_):
            d0 = self.main.wait_for('HQ_HASH_NEW')
            f = file(self.root + '/a_file', 'w')
            print >> f, 'hola'
            os.unlink(self.root + '/a_file')
            f.close()
            print >> file(self.root + '/b_file', 'w'), 'chau'
            return d0
        d = self.main.wait_for('SYS_LOCAL_RESCAN_DONE')
        self.main.start()
        d.addCallback(cb)
        return d

    def test_handle_AQ_DOWNLOAD_DOES_NOT_EXIST(self):
        """handle_AQ_DOWNLOAD_DOES_NOT_EXIST."""
        sync = Sync(main=self.main)
        self.called = False

        def faked_nothing(ssmr, event, params, *args):
            """Wrap SSMR.nothing to test."""
            self.called = True
        self.patch(SyncStateMachineRunner, 'nothing', faked_nothing)

        kwargs = dict(share_id='share_id', node_id='node_id')
        sync.handle_AQ_DOWNLOAD_DOES_NOT_EXIST(**kwargs)
        self.assertTrue(self.called, 'nothing was called')

    def test_handle_FILE_CREATE_while_LOCAL(self):
        """A FS_FILE_CREATE is received with the node in LOCAL."""
        sync = Sync(main=self.main)
        self.called = False

        def faked_nothing(ssmr, event, params, *args):
            """Wrap SSMR.nothing to test."""
            self.called = True
        self.patch(SyncStateMachineRunner, 'nothing', faked_nothing)

        # create a file and put it in local
        fsm = self.main.fs
        somepath = os.path.join(self.root, 'somepath')
        mdid = fsm.create(somepath, '')
        fsm.set_by_mdid(mdid, local_hash='somehash')

        # send the event, and check that it called its .nothing()
        sync.handle_FS_FILE_CREATE(somepath)
        self.assertTrue(self.called)

    def test_SV_HASH_NEW_with_file_uploadinterrupted(self):
        """A SV_HASH_NEW is received after upload interrupted."""
        sync = Sync(main=self.main)
        self.called = False

        def fake_meth(_, event, params, hash):
            """Wrap SSMR.reput_file_from_local to test."""
            self.assertEqual(event, 'SV_HASH_NEW')
            self.assertEqual(hash, '')
            self.called = True
        self.patch(SyncStateMachineRunner, 'reput_file_from_local', fake_meth)

        # create a file and put it in local, without server_hash, as
        # if the upload was cut in the middle after the make file
        fsm = self.main.fs
        somepath = os.path.join(self.root, 'somepath')
        mdid = fsm.create(somepath, '', node_id='node_id')
        fsm.set_by_mdid(mdid, local_hash='somehash', crc32='crc32',
                        stat='stat', size='size')

        # send the event and check
        mdobj = fsm.get_by_mdid(mdid)
        sync._handle_SV_HASH_NEW(mdobj.share_id, mdobj.node_id, '') # no content
        self.assertTrue(self.called)


class SyncStateMachineRunnerTestCase(BaseSync):
    """Tests for the SyncStateMachineRunner."""

    def setUp(self):
        """Init."""
        BaseSync.setUp(self)

        # create a file
        self.fsm = self.main.fs
        somepath = os.path.join(self.root, 'somepath')
        self.mdid = self.fsm.create(somepath, '', node_id='node_id')

        key = FSKey(self.main.fs, share_id='', node_id='node_id')
        self.ssmr = SyncStateMachineRunner(fsm=self.main.fs, main=self.main,
                                           key=key, logger=None)
        self.root = self.mktemp('root')

    def tearDown(self):
        """Clean up."""
        self.ssmr = None
        BaseSync.tearDown(self)

    def test_delete_file_without_hash(self):
        """Delete_file can be called without the server hash."""
        self.ssmr.delete_file(event='AQ_DOWNLOAD_ERROR', params=None)

    def test_delete_file_with_hash(self):
        """Delete_file can be called with the server hash."""
        self.ssmr.delete_file(event='AQ_DOWNLOAD_ERROR', params=None,
                              server_hash='')

    def test_validateactualdata_log(self):
        """This method should log detailed info."""
        # create testing data
        somepath = os.path.join(self.root, 'somepath')
        open(somepath, "w").close()
        oldstat = os.stat(somepath)
        f = open(somepath, "w")
        f.write("new")
        f.close()

        # log config
        handler = testcase.MementoHandler()
        handler.setLevel(logging.DEBUG)
        self.ssmr.log.addHandler(handler)

        # call the method with args to not match
        self.ssmr.validate_actual_data(somepath, oldstat)

        # check log
        self.assertEqual(len(handler.records), 1)
        log_msg = handler.records[0].message
        self.assertTrue("st_ino" in log_msg)
        self.assertTrue("st_size" in log_msg)
        self.assertTrue("st_mtime" in log_msg)
        self.ssmr.log.removeHandler(handler)

    def test_put_file_stores_info(self):
        """The put_file method should store the info in FSM."""
        self.ssmr.put_file('HQ_HASH_NEW', None, 'hash', 'crc', 'size', 'stt')

        # check the info is stored
        mdobj = self.fsm.get_by_mdid(self.mdid)
        self.assertEqual(mdobj.local_hash, 'hash')
        self.assertEqual(mdobj.crc32, 'crc')
        self.assertEqual(mdobj.size, 'size')
        self.assertEqual(mdobj.stat, 'stt')

    def test_reput_file_stores_info(self):
        """The reput_file method should store the info in FSM."""
        self.ssmr.reput_file('HQ_HASH_NEW', None, 'hash', 'crc', 'size', 'stt')

        # check the info is stored
        mdobj = self.fsm.get_by_mdid(self.mdid)
        self.assertEqual(mdobj.local_hash, 'hash')
        self.assertEqual(mdobj.crc32, 'crc')
        self.assertEqual(mdobj.size, 'size')
        self.assertEqual(mdobj.stat, 'stt')


class FakedState(object):
    """A faked state."""

    def __init__(self, action_func):
        self.action_func = action_func
        self.values = []

    def get_transition(self, event_name, parameters):
        """A fake get_transition."""
        class A(object): pass
        result = A()
        result.action_func = self.action_func
        result.target = self.values
        return result


class StateMachineRunnerTestCase(BaseTwistedTestCase):
    """Test suite for StateMachineRunner."""

    def setUp(self):
        """Init."""
        BaseTwistedTestCase.setUp(self)

        logger = logging.getLogger('ubuntuone.SyncDaemon.sync')
        self.smr = fsm_module.StateMachineRunner(None, logger)
        self.handler = testcase.MementoHandler()
        self.handler.setLevel(logging.INFO)
        logger.addHandler(self.handler)

    def test_on_event_logs_info(self):
        """Test proper logging in on_event."""
        event_name = 'TEST_EVENT'
        parameters = object()
        action_func = 'SOME_ACTION_FUNC'
        setattr(self.smr, action_func, lambda *_: None)
        self.patch(self.smr, 'get_state', lambda: FakedState(action_func))
        self.smr.on_event(event_name=event_name, parameters=parameters)

        # assert over logging
        self.assertTrue(len(self.handler.records), 1)

        record = self.handler.records[0]
        self.assertEqual(record.levelno, logging.INFO)

        msg = record.message
        error = '%s must be in record.message (%s)'
        self.assertTrue(event_name in msg, error % (event_name, msg))
        self.assertTrue(str(parameters) in msg, error % (parameters, msg))
        self.assertTrue(action_func in msg, error % (action_func, msg))


class TestNewGenerationOnOperations(BaseSync):
    """Test we handle ok the new generation in some operations."""

    def setUp(self):
        """Set up."""
        super(TestNewGenerationOnOperations, self).setUp()
        self.sync = Sync(main=self.main)
        self.handler.setLevel(logging.DEBUG)

        key = FSKey(self.main.fs, share_id='', node_id='node_id')
        self.ssmr = SyncStateMachineRunner(fsm=self.main.fs, main=self.main,
                                           key=key, logger=None)
        self.vm = self.main.vm

    def test_handle_AQ_UNLINK_OK(self):
        """Test that AQ_UNLINK_OK calls the generation handler."""
        called = []
        self.patch(SyncStateMachineRunner, 'update_generation',
                   lambda s, *a: called.append(a))

        d = dict(share_id='volume_id', node_id='node_id', parent_id='parent',
                 new_generation=77)
        self.sync.handle_AQ_UNLINK_OK(**d)
        self.assertEqual(called, [('volume_id', "node_id", 77)])

    def test_handle_AQ_MOVE_OK(self):
        """Test that AQ_MOVE_OK calls the generation handler."""
        called = []
        self.patch(SyncStateMachineRunner, 'update_generation',
                   lambda s, *a: called.append(a))

        d = dict(share_id='volume_id', node_id='node_id', new_generation=32)
        self.sync.handle_AQ_MOVE_OK(**d)
        self.assertEqual(called, [('volume_id', "node_id", 32)])

    def test_handle_AQ_UPLOAD_FINISHED(self):
        """Test that AQ_UPLOAD_FINISHED calls the generation handler."""
        called = []
        self.patch(SyncStateMachineRunner, 'update_generation',
                   lambda s, *a: called.append(a))

        d = dict(share_id='volume_id', node_id='node_id',
                 hash='hash', new_generation=15)
        self.sync.handle_AQ_UPLOAD_FINISHED(**d)
        self.assertEqual(called, [('volume_id', "node_id", 15)])

    def test_handle_AQ_FILE_NEW_OK(self):
        """Test that AQ_FILE_NEW_OK calls the generation handler."""
        called = []
        self.patch(SyncStateMachineRunner, 'update_generation',
                   lambda s, *a: called.append(a))

        d = dict(marker='mdid', new_id='new_id', new_generation=12,
                 volume_id=ROOT)
        self.sync.handle_AQ_FILE_NEW_OK(**d)
        self.assertEqual(called, [(ROOT, "new_id", 12)])

    def test_handle_AQ_DIR_NEW_OK(self):
        """Test that AQ_DIR_NEW_OK calls the generation handler."""
        called = []
        self.patch(SyncStateMachineRunner, 'update_generation',
                   lambda s, *a: called.append(a))

        d = dict(marker='mdid', new_id='new_id', new_generation=17,
                 volume_id=ROOT)
        self.sync.handle_AQ_DIR_NEW_OK(**d)
        self.assertEqual(called, [(ROOT, "new_id", 17)])

    def test_checknewvol_no_volume(self):
        """Log warning if volume does not exist."""
        not_existant_vol = str(uuid.uuid4())
        self.ssmr.update_generation(not_existant_vol, "node_id", 77)
        self.assertTrue(self.handler.check_warning('Volume not found'))

    def test_checknewvol_smaller_gen(self):
        """Only log debug if new generation smaller than current."""
        self.vm.update_generation(ROOT, 15)
        self.ssmr.update_generation(ROOT, "node_id", 14)
        self.assertTrue(self.handler.check_info(
                        'Got smaller or equal generation'))

    def test_checknewvol_same_gen(self):
        """Only log debug if new generation equal than current."""
        self.vm.update_generation(ROOT, 15)
        self.ssmr.update_generation(ROOT, "node_id", 15)
        self.assertTrue(self.handler.check_info(
                        'Got smaller or equal generation'))

    def test_checknewvol_gen_current_plus_one(self):
        """Set new volume generation if current plus one."""
        self.vm.update_generation(ROOT, 15)
        self.ssmr.update_generation(ROOT, "node_id", 16)
        self.assertEqual(self.vm.get_volume(ROOT).generation, 16)
        self.assertTrue(self.handler.check_info('Updating current generation'))

    def test_checknewvol_lot_bigger(self):
        """Ask for new delta if new generation is much bigger than current."""
        # set up
        called = []
        self.main.action_q.get_delta = lambda *a: called.append(a)
        self.vm.update_generation(ROOT, 15)

        # call the method
        self.ssmr.update_generation(ROOT, "node_id", 17)

        # check that generation didn't change, we asked for delta, and logged
        self.assertEqual(self.vm.get_volume(ROOT).generation, 15)
        self.assertEqual(called, [(ROOT, 15)])
        self.assertTrue(self.handler.check_info('Generation much bigger'))

    def test_checknewvol_new_gen_is_None(self):
        """Log warning if volume does not exist."""
        self.vm.update_generation(ROOT, 1)
        self.ssmr.update_generation(ROOT, "node_id", None)
        self.assertTrue(self.handler.check_debug(
                        'Client not ready for generations'))

    def test_checknewvol_volume_gen_is_None(self):
        """Log warning if volume does not exist."""
        assert self.vm.get_volume(ROOT).generation is None
        self.ssmr.update_generation(ROOT, "node_id", 15)
        self.assertTrue(self.handler.check_debug(
                        'Client not ready for generations'))

    def test_check_generation_on_node_set(self):
        """Check that we update the generation of the node."""
        # create the fake file
        self.main.vm._got_root("parent_id")
        self.sync._handle_SV_FILE_NEW(ROOT, "node_id", "parent_id", "file")

        # update generation
        self.ssmr.update_generation(ROOT, "node_id", 15)

        # test
        node = self.main.fs.get_by_node_id(ROOT, "node_id")
        self.assertEqual(node.generation, 15)

    def test_check_generation_on_node_set_wont_fail(self):
        """Check that if there is no node we dont fail."""
        # update generation
        self.ssmr.update_generation(ROOT, "node_id", 15)

    def test_save_generation_after_seting_node_id(self):
        """Test that we call update_generation after the ssmr handler."""
        root_id = uuid.uuid4()
        self.main.vm._got_root(root_id)
        mdobj = self.main.fs.get_by_node_id(ROOT, root_id)
        path = os.path.join(
            self.main.fs.get_abspath(ROOT, mdobj.path), "file")
        self.main.fs.create(path=path, share_id=ROOT, is_dir=False)
        node = self.main.fs.get_by_path(path)
        d = dict(marker=MDMarker(node.mdid),
                 new_id='new_id', new_generation=12,
                 volume_id=ROOT)
        self.sync.handle_AQ_FILE_NEW_OK(**d)

        # test
        node = self.main.fs.get_by_node_id(ROOT, "new_id")
        self.assertEqual(node.generation, 12)

class TestHandleAqDeltaOk(BaseSync):
    """Sync.handle_AQ_DELTA_OK handles the recepcion of a new delta and applies
    all the changes that came from it."""

    def setUp(self):
        """Do the setUp."""

        super(TestHandleAqDeltaOk, self).setUp()
        self.sync = Sync(main=self.main)
        root_id = uuid.uuid4()
        self.main.vm._got_root(root_id)

        self.filetxtdelta = delta.FileInfoDelta(
            generation=5, is_live=True, file_type=delta.FILE,
            parent_id=root_id, share_id=ROOT, node_id=uuid.uuid4(),
            name="file.txt", is_public=False, content_hash="hash",
            crc32=1, size=10, last_modified=0)

        self.dirdelta = delta.FileInfoDelta(
            generation=6, is_live=True, file_type=delta.DIRECTORY,
            parent_id=root_id, share_id=ROOT, node_id=uuid.uuid4(),
            name="directory", is_public=False, content_hash="hash",
            crc32=1, size=10, last_modified=0)

    def create_filetxt(self):
        """Create a file based on self.filetxtdelta."""

        dt = self.filetxtdelta
        mdobj = self.main.fs.get_by_node_id(dt.share_id, dt.parent_id)
        path = os.path.join(
            self.main.fs.get_abspath(dt.share_id, mdobj.path), dt.name)
        self.main.fs.create(
            path=path, share_id=dt.share_id, node_id=dt.node_id,
            is_dir=False)
        node = self.main.fs.get_by_node_id(dt.share_id, dt.node_id)
        self.main.fs.set_by_mdid(node.mdid, generation=dt.generation)

    def create_dir(self):
        """Create a directory based on self.dirdelta."""

        dt = self.dirdelta
        mdobj = self.main.fs.get_by_node_id(dt.share_id, dt.parent_id)
        path = os.path.join(
            self.main.fs.get_abspath(dt.share_id, mdobj.path), dt.name)
        self.main.fs.create(
            path=path, share_id=dt.share_id, node_id=dt.node_id,
            is_dir=True)
        node = self.main.fs.get_by_node_id(dt.share_id, dt.node_id)
        self.main.fs.set_by_mdid(node.mdid, generation=dt.generation)

    def test_not_full(self):
        """If we dont have a full delta, we need to ask for another one."""
        sync = Sync(main=self.main)
        called = []
        self.main.action_q.get_delta = lambda *a: called.append(a)

        kwargs = dict(volume_id=ROOT, delta_content=[], end_generation=11,
                      full=False, free_bytes=0)
        sync.handle_AQ_DELTA_OK(**kwargs)

        self.assertEqual(called, [(ROOT, 11)])

    def test_free_bytes_set(self):
        """The volume gets the free bytes set."""
        sync = Sync(main=self.main)

        kwargs = dict(volume_id=ROOT, delta_content=[], end_generation=11,
                      full=True, free_bytes=10)
        sync.handle_AQ_DELTA_OK(**kwargs)

        self.assertEqual(self.main.vm.get_volume(ROOT).free_bytes, 10)

    def test_end_generation_set(self):
        """The volume gets the end generation set."""
        sync = Sync(main=self.main)

        kwargs = dict(volume_id=ROOT, delta_content=[], end_generation=11,
                      full=True, free_bytes=10)
        sync.handle_AQ_DELTA_OK(**kwargs)

        self.assertEqual(self.main.vm.get_volume(ROOT).generation, 11)

    def test_node_generation_older_skip(self):
        """The node does not get the new generation set."""
        self.create_filetxt()

        dt2 = copy.copy(self.filetxtdelta)
        dt2.generation = self.filetxtdelta.generation - 1
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        node = self.main.fs.get_by_node_id(ROOT, self.filetxtdelta.node_id)
        self.assertEqual(node.generation, self.filetxtdelta.generation)

    def test_new_file(self):
        """Make sure a live file in the delta is in fs after executed."""
        deltas = [ self.filetxtdelta ]

        kwargs = dict(volume_id=ROOT, delta_content=deltas, end_generation=11,
                      full=True, free_bytes=10)
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the file is created
        node = self.main.fs.get_by_node_id(ROOT, self.filetxtdelta.node_id)
        self.assertEqual(node.path, self.filetxtdelta.name)
        self.assertEqual(node.is_dir, False)
        self.assertEqual(node.generation, self.filetxtdelta.generation)

    def test_existing_file_still_there(self):
        """A file will still exist after a delta arrives."""
        self.create_filetxt()

        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.generation = 8
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the file is created
        node = self.main.fs.get_by_node_id(ROOT, self.filetxtdelta.node_id)
        self.assertEqual(node.generation, dt2.generation)

    def test_existing_file_dead(self):
        """The handler for SV_FILE_DELETED is called"""
        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.is_live = False
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        called = []
        self.sync._handle_SV_FILE_DELETED = \
                   lambda *args, **kwargs: called.append((args, kwargs))
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the handler is created
        self.assertTrue(called)

    def test_new_dir(self):
        """Make sure a live dir in the delta is in fs after executed."""

        deltas = [ self.dirdelta ]

        kwargs = dict(volume_id=ROOT, delta_content=deltas, end_generation=11,
                      full=True, free_bytes=10)
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the dir is created
        node = self.main.fs.get_by_node_id(ROOT, self.dirdelta.node_id)
        self.assertEqual(node.path, self.dirdelta.name)
        self.assertEqual(node.is_dir, True)
        self.assertEqual(node.generation, self.dirdelta.generation)

    def test_sv_hash_new_called_for_file(self):
        """The handler for SV_HASH_NEW is called"""
        self.create_filetxt()

        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.generation = self.filetxtdelta.generation + 1
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        called = []
        self.sync._handle_SV_HASH_NEW = \
                   lambda *args, **kwargs: called.append((args, kwargs))
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the handler is created
        self.assertTrue(called)

    def test_sv_hash_new_not_called_for_dir(self):
        """The handler for SV_HASH_NEW is not called"""
        self.create_dir()

        # send a new delta
        dt2 = copy.copy(self.dirdelta)
        dt2.generation = self.dirdelta.generation + 1
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        called = []
        self.sync._handle_SV_HASH_NEW = \
                   lambda *args, **kwargs: called.append((args, kwargs))
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the handler is created
        self.assertFalse(called)

    def test_sv_moved_called(self):
        """The handler for SV_MOVED is called"""
        self.create_dir()
        self.create_filetxt()

        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.generation = self.dirdelta.generation + 1
        dt2.parent_id = self.dirdelta.node_id
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        called = []
        self.sync._handle_SV_MOVED = \
                   lambda *args, **kwargs: called.append((args, kwargs))
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the handler is created
        self.assertTrue(called)

    def test_sv_moved_called_name(self):
        """The handler for SV_MOVED is called"""
        self.create_dir()
        self.create_filetxt()

        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.generation = self.dirdelta.generation + 1
        dt2.name = "newname"
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        called = []
        self.sync._handle_SV_MOVED = \
                   lambda *args, **kwargs: called.append((args, kwargs))
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the handler is created
        self.assertTrue(called)

    def test_sv_moved_not_called(self):
        """The handler for SV_MOVED is not called"""
        self.create_dir()
        self.create_filetxt()

        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.generation = self.dirdelta.generation + 1
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        called = []
        self.sync._handle_SV_MOVED = \
                   lambda *args, **kwargs: called.append((args, kwargs))
        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check that the handler is created
        self.assertFalse(called)

    def test_exception_logged(self):
        """We call self.logger.exception on error."""
        # send a new delta
        dt2 = copy.copy(self.filetxtdelta)
        dt2.is_live = False
        kwargs = dict(volume_id=ROOT, delta_content=[dt2], end_generation=11,
                      full=True, free_bytes=10)
        self.sync._handle_SV_FILE_DELETED = \
                   lambda *args, **kwargs: 1/0
        handler = testcase.MementoHandler()
        handler.setLevel(logging.ERROR)
        self.sync.logger.addHandler(handler)

        self.sync.handle_AQ_DELTA_OK(**kwargs)

        # check log
        self.assertEqual(len(handler.records), 1)
        log_msg = handler.records[0].message
        self.assertTrue("can't be applied." in log_msg)
        self.sync.logger.removeHandler(handler)
        self.handler.records = []
